package fetch

import (
	. "cf_index/common/util"
	. "cf_index/util"
	"fmt"
	"github.com/Shopify/sarama"
	"github.com/go-ini/ini"
	"log"
	"os"
	"strings"
	"sync"
)

func init() {
	sarama.Logger = log.New(os.Stdout, "[Sarama] ", log.LstdFlags)
}

type nativeConsumer struct {
	BrokerList   string
	Offset       string
	Topics       []string
	topicAttr    map[string]string //topic attr
	consumer_sli []*consumer
	running      bool //status
}

var (
	natconsumer         *nativeConsumer
	once_nativeConsumer sync.Once
)

func NativeConsumer() *nativeConsumer {
	once_nativeConsumer.Do(func() {
		natconsumer = &nativeConsumer{
			topicAttr: make(map[string]string),
		}
		if err := natconsumer.Init(); err != nil {
			log.Fatalln(err)
		}
		if err := natconsumer.Start(); err != nil {
			log.Fatalln(err)
		}
		natconsumer.running = true
	})
	return natconsumer
}

func (n *nativeConsumer) Init() error {
	conffile := ConfFile()
	param, err := ini.Load(conffile.File)
	if err != nil {
		return err
	}

	//1
	n.BrokerList = param.Section("kafka").Key("hosts").String()
	n.Offset = param.Section("kafka").Key("where").String()

	//2 首页根绝航线,挖掘特价日期和价格,一天更新一次
	mp_autoupdate_topic := param.Section("kafka").Key("mp_autoupdate_topic").String()
	//2.1
	n.topicAttr[mp_autoupdate_topic] = "mp_autoupdate_topic"
	n.Topics = append(n.Topics, mp_autoupdate_topic)

	//3 读取爬虫国际航线数据
	inter_topics_file := param.Section("kafka").Key("inter_topics").String()
	inter_compact_topics_file := param.Section("kafka").Key("inter_compact_topics").String()
	inter_qy_topics_file := param.Section("kafka").Key("inter_qy_topics").String()
	inter_topics_new_file := param.Section("kafka").Key("inter_topics_new").String()
	inter_update_topic := param.Section("kafka").Key("inter_update_topic").String()
	//3.1
	n.ReadFile(inter_topics_file, "inter_topics")
	n.ReadFile(inter_compact_topics_file, "inter_compact_topics")
	n.ReadFile(inter_qy_topics_file, "inter_qy_topics")
	n.ReadFile(inter_topics_new_file, "inter_topics_new")
	n.topicAttr[inter_update_topic] = "inter_update_topic"
	n.Topics = append(n.Topics, inter_update_topic)

	//4 国际特价批量挖掘
	inter_exploit_topic := param.Section("kafka").Key("inter_exploit_topic").String()
	inter_exploit_update_topic := param.Section("kafka").Key("inter_exploit_update_topic").String()
	//4.1
	n.topicAttr[inter_exploit_topic] = "inter_exploit_topic"
	n.Topics = append(n.Topics, inter_exploit_topic)
	n.topicAttr[inter_exploit_update_topic] = "inter_exploit_update_topic"
	n.Topics = append(n.Topics, inter_exploit_update_topic)

	//5 读取国内航线数据
	china_crawl_topic_file := param.Section("kafka").Key("china_crawl_topics").String()
	china_crawl_compact_topic_file := param.Section("kafka").Key("china_crawl_compact_topics").String()
	china_topic := param.Section("kafka").Key("china_topic").String()
	china_update_topic := param.Section("kafka").Key("china_update_topic").String()
	//5.1
	n.ReadFile(china_crawl_topic_file, "china_crawl_topics")
	n.ReadFile(china_crawl_compact_topic_file, "china_crawl_compact_topics")
	n.topicAttr[china_topic] = "china_topic"
	n.Topics = append(n.Topics, china_topic)
	n.topicAttr[china_update_topic] = "china_update_topic"
	n.Topics = append(n.Topics, china_update_topic)

	//6 读取更新日志信息
	islog_topic := param.Section("kafka").Key("islog_topic").String()
	iclog_topic := param.Section("kafka").Key("iclog_topic").String()
	//6.1
	n.topicAttr[islog_topic] = "islog_topic"
	n.Topics = append(n.Topics, islog_topic)
	n.topicAttr[iclog_topic] = "iclog_topic"
	n.Topics = append(n.Topics, iclog_topic)

	return nil
}

func (n *nativeConsumer) ReadFile(file, attr string) {
	if sli, err := ReadFromFile(file); err == nil {
		for _, topic := range sli {
			n.topicAttr[topic] = attr
		}
		n.Topics = append(n.Topics, sli...)
	}
}

func (n *nativeConsumer) Start() error {
	if n.running {
		return nil
	}
	for _, topic := range n.Topics {
		c := NewConsumer(n, topic, n.Offset)
		if err := c.Start(); err == nil {
			n.consumer_sli = append(n.consumer_sli, c)
		}
	}
	return nil
}

func (n *nativeConsumer) Stop() {
	for _, c := range n.consumer_sli {
		c.Stop()
	}
	n.running = false
}

//--------------------------------------------------------------------------------------------------
//--------------------------------------------------------------------------------------------------
type consumer struct {
	native   *nativeConsumer
	topic    string
	offset   string
	consumer sarama.Consumer
	messages chan *sarama.ConsumerMessage
	closing  chan struct{}  //close consumer
	wg       sync.WaitGroup //wait for consumer  closed
}

func NewConsumer(n *nativeConsumer, topic, offset string) *consumer {
	return &consumer{
		native:   n,
		topic:    topic,
		offset:   offset,
		messages: make(chan *sarama.ConsumerMessage, bufferSize),
		closing:  make(chan struct{}),
	}
}

func (c *consumer) Start() error {
	var initialOffset int64
	switch c.offset {
	case "begin":
		initialOffset = sarama.OffsetOldest
	case "now":
		initialOffset = sarama.OffsetNewest
	default:
		return fmt.Errorf("[NativeConsumer] offset should be `begin` or `now`")
	}
	myconsumer, err := sarama.NewConsumer(strings.Split(c.native.BrokerList, ","), nil)
	if err != nil {
		return fmt.Errorf("[NativeConsumer] Failed to start consumer, %s", err.Error())
	}
	c.consumer = myconsumer
	partitionList, err := c.GetPartitions()
	if err != nil {
		return fmt.Errorf("[NativeConsumer] Failed to get the list of partition,%s", err.Error())
	}

	for _, partition := range partitionList {
		pc, err := c.consumer.ConsumePartition(c.topic, partition, initialOffset)
		if err != nil {
			return fmt.Errorf("[NativeConsumer] Failed to start consumer for partiton: %d: %s", partition, err.Error())
		}
		go func(pc sarama.PartitionConsumer) {
			<-c.closing
			//pc.AsyncClose()
			if err := pc.Close(); err != nil {
				log.Printf("[NativeConsumer] Closing partition consumer of %v fail, %s\n", c.topic, err.Error())
			}
		}(pc)

		c.wg.Add(1)
		go func(pc sarama.PartitionConsumer) {
			defer c.wg.Done()
			for message := range pc.Messages() {
				c.messages <- message //message如果不被消费,关闭pc.Messages()时,退出时会卡住,会导致c.wg.Done不被调用
			}
		}(pc)

		go func(attr string, c *consumer) {
			FindBuilder(attr, c.messages)
			//for msg := range c.messages {
			//	log.Printf("%v %v %v %v\n", msg.Topic, msg.Partition, msg.Offset,
			//		string(msg.Key),
			//	)
			//}
		}(c.native.topicAttr[c.topic], c)
	}
	log.Printf("[NativeConsumer] Successful start of consumer %s\n", c.topic)
	return nil
}

func (c *consumer) Stop() {
	log.Println("[NativeConsumer] Initiating shutdown of consumer of", c.topic)
	close(c.closing)
	//如果topic对应的chan 的消息不被消费,当关闭topic对应chan通道时,会导致关闭信息不能被读取到,卡在c.wg.Wait处
	c.wg.Wait()
	close(c.messages)
	log.Printf("[NativeConsumer] Done consuming topic %s", c.topic)

	if err := c.consumer.Close(); err != nil {
		log.Println("[NativeConsumer] Failed to close consumer: ", err)
	}
}

func (c *consumer) GetPartitions() (pList []int32, err error) {
	return c.consumer.Partitions(c.topic)
}
